static signed char __Numba_PyInt_AsSignedChar(PyObject* x);
static unsigned char __Numba_PyInt_AsUnsignedChar(PyObject* x);
static signed short __Numba_PyInt_AsSignedShort(PyObject* x);
static unsigned short __Numba_PyInt_AsUnsignedShort(PyObject* x);
static signed int __Numba_PyInt_AsSignedInt(PyObject* x);
static unsigned int __Numba_PyInt_AsUnsignedInt(PyObject* x);
static signed long __Numba_PyInt_AsSignedLong(PyObject* x);
static unsigned long __Numba_PyInt_AsUnsignedLong(PyObject* x);
static signed PY_LONG_LONG __Numba_PyInt_AsSignedLongLong(PyObject* x);
static unsigned PY_LONG_LONG __Numba_PyInt_AsUnsignedLongLong(PyObject* x);
